import ast


class FileToResourceTransform(ast.NodeTransformer):

    # pylint: disable=pylint(invalid-name)
    def visit_Name(self, node: ast.Name):
        node = self.generic_visit(node)

        if isinstance(node.ctx, ast.Load) and node.id == '__file__':
            return ast.Name('__resource__', ast.Load())

        return node


def file_to_resource_transform(source: str):
    tree = ast.parse(source, type_comments=True)
    tree = FileToResourceTransform().visit(tree)
    tree = ast.fix_missing_locations(tree)
    return ast.unparse(tree)



from collections import OrderedDict
from os import makedirs, walk
from pathlib import Path
from shutil import copy
from tempfile import mkdtemp
from typing import Dict, List

from pdistx.utils.path import fnmatch_any, rmpath
from pdistx.utils.source import read_source
from pdistx.utils.zip import zipit

from .checks import has_absolute_import_of_module, has_relative_import
from .transform import file_to_resource_transform


def perform(
    source: Path,
    target: Path,
    filters: List[Path],
    resources: bool,
    main: bool,
    zip_: Path,
):
    # ensure pre-conditions
    assert source.is_dir(), 'source is expected to be a directory'

    if zip_:
        assert not target.is_absolute(), 'target path is expected to be relative'

    # list of temporary files and folders
    tmps: List[Path] = []

    # temporary paths get cleaned automatically at the end of this block
    try:

        # purging target or zip
        if zip_:
            print(f'Purging {zip_}...')
            rmpath(zip_)
        else:
            print(f'Purging {target}...')
            rmpath(target)

        # determine base directory and relative target path
        if zip_:
            intermediate = Path(mkdtemp())
            tmps.append(intermediate)
        else:
            intermediate = target.parent
            target = Path(target.name)

        # prepare output paths
        packed = intermediate.joinpath(target.parent, target)

        if resources:
            resources_root = intermediate.joinpath(target.parent, target.stem + '_resources')
            print(f'Purging {resources_root}...')
            rmpath(resources_root)

        # process all files
        modules: Dict[str, (str, bool)] = {}

        for source_folder, folders, files in walk(source, followlinks=True):

            # prepare folders
            source_folder = Path(source_folder)
            package_folder = source_folder.relative_to(source)

            if resources:
                resource_folder = resources_root.joinpath(package_folder)

            # filter entries to be ignored (folders need to be modified in-place to take effect for os.walk)
            def _folder_filter(folder: Path):
                return not fnmatch_any(folder.name, ['__pycache__', '.git']) and folder not in filters

            def _file_filter(file: Path):
                return not fnmatch_any(file.name, ['*.pyc']) and file not in filters

            folders[:] = [folder for folder in folders if _folder_filter(source_folder.joinpath(folder))]
            files = [file for file in files if _file_filter(source_folder.joinpath(file))]

            # read or copy file
            for file in files:
                source_file = source_folder.joinpath(file)

                # read module codes
                if file.endswith('.py'):
                    # determine module name
                    is_package = file == '__init__.py'

                    name = list(package_folder.parts)
                    name += [file.split('.')[0]] if not is_package else []

                    name = '.'.join(name)

                    # load module code
                    code = read_source(source_file)

                    if resources:
                        code = file_to_resource_transform(code)

                    # check code for invalid imports
                    if name == '__main__' and has_relative_import(code):
                        raise ValueError(f'{source_file} contains a relative import, which is forbidden')

                    if name != '__main__' and has_absolute_import_of_module(code, source.name):
                        raise ValueError(
                            f'{source_file} contains an absolute import of {source.name}, which is forbidden')

                    # assign to module dictionay
                    modules[name] = (code, is_package)

                # copy resource files
                else:
                    if resources:
                        makedirs(resource_folder, exist_ok=True)
                        copy(source_file, resource_folder.joinpath(file), follow_symlinks=True)

        if len(modules) == 0:
            raise ValueError('no modules found')

        # create all missing intermediate packages
        for name in list(modules.keys()):
            parts = name.split('.')
            for i in range(0, len(parts)):
                package = '.'.join(parts[0:i])
                if package not in modules:
                    modules[package] = ('', True)

        # ensure stable ordering
        modules = OrderedDict(sorted(modules.items(), key=lambda i: i[0]))

        # determine bootstrap module
        if main:
            mode = 'main'
            bootstrap, _ = modules.get('__main__', (None, None))
        else:
            mode = 'package'
            bootstrap, _ = modules.get('', (None, None))

        if bootstrap is None:
            raise RuntimeError('bootstrap module is missing')

        # create packed file
        code = read_source(Path(__file__).parent.joinpath('template.py')).split('\n')

        injected_mode = 0
        injected_name = 0
        injected_modules = 0

        for i in range(len(code)):
            if '    pack_mode = \'\'' == code[i]:
                code[i] = '    pack_mode = ' + repr(mode)
                injected_mode += 1
            elif '    pack_name = \'\'' == code[i]:
                code[i] = '    pack_name = ' + repr(source.name)
                injected_name += 1
            elif '    pack_modules = OrderedDict()' == code[i]:
                code[i] = '    pack_modules = ' + repr(modules)
                injected_modules += 1

        if injected_mode != 1 or injected_name != 1 or injected_modules != 1:
            raise RuntimeError('inconsistent code template')

        code = '\n'.join(code) + '\n\n' + bootstrap

        print(f'Writing {packed}...')
        with open(packed, 'w', encoding='utf-8') as file:
            file.write(code)

        # zip intermediate path to zip path
        if zip_:
            print(f'Packing {zip_}...')
            zipit(intermediate, zip_, Path(''))

    finally:
        # clean up temporary folders
        for path in tmps:
            print(f'Purging {path}...')
            rmpath(path)



import ast
from pathlib import Path
from typing import List

from pdistx.utils.source import read_source


class _ImportNameStringTransform(ast.NodeTransformer):

    def __init__(self, level, modules):
        self._level = level
        self._modules = modules
        self.uses_package_or_name = False
        self.string_rewrite_applied = False
        super().__init__()

    # pylint: disable=pylint(invalid-name)
    def visit_Name(self, node: ast.Name):
        node = self.generic_visit(node)

        if node.id in ['__package__', '__name__']:
            self.uses_package_or_name = True

        return node

    # pylint: disable=pylint(invalid-name)
    def visit_Constant(self, node: ast.Constant):
        node = self.generic_visit(node)

        if isinstance(node.value, str):
            if node.value.split('.')[0] in self._modules:
                self.string_rewrite_applied = True
                return ast.Call(
                    func=ast.Attribute(value=ast.Constant(value='.'), attr='join', ctx=ast.Load()),
                    args=[
                        ast.BinOp(
                            left=ast.Subscript(
                                value=ast.Call(
                                    func=ast.Attribute(
                                        value=ast.Name(id='__package__', ctx=ast.Load()),
                                        attr='split',
                                        ctx=ast.Load(),
                                    ),
                                    args=[ast.Constant(value='.')],
                                    keywords=[],
                                ),
                                slice=ast.Slice(upper=ast.UnaryOp(
                                    op=ast.USub(),
                                    operand=ast.Constant(value=self._level),
                                )),
                                ctx=ast.Load(),
                            ),
                            op=ast.Add(),
                            right=ast.List(elts=[node], ctx=ast.Load()),
                        )
                    ],
                    keywords=[],
                )

        return node


def _transform_import_name_string(level, modules, name):
    visitor = _ImportNameStringTransform(level, modules)
    name = visitor.visit(name)
    if visitor.uses_package_or_name or not visitor.string_rewrite_applied:
        return None
    return name


class ImportTransform(ast.NodeTransformer):

    def __init__(self, level, modules):
        self._level = level
        self._modules = modules
        super().__init__()

    # pylint: disable=pylint(invalid-name)
    def visit_Import(self, node: ast.Import):
        node = self.generic_visit(node)

        nodes = []
        for name in node.names:
            # Keep "import abc.def" and "import abc.def as xyz" for non included modules
            if name.name.split('.')[0] not in self._modules:
                nodes.append(ast.Import([name]))
                continue

            # Perform an absolute import with "__import__" to ensure nested imports are solved properly
            nodes.append(
                ast.Expr(value=ast.Call(
                    func=ast.Name(id='__import__', ctx=ast.Load()),
                    args=[
                        ast.Call(
                            func=ast.Attribute(value=ast.Constant(value='.'), attr='join', ctx=ast.Load()),
                            args=[
                                ast.BinOp(
                                    left=ast.Subscript(
                                        value=ast.Call(
                                            func=ast.Attribute(
                                                value=ast.Name(id='__package__', ctx=ast.Load()),
                                                attr='split',
                                                ctx=ast.Load(),
                                            ),
                                            args=[ast.Constant(value='.')],
                                            keywords=[],
                                        ),
                                        slice=ast.Slice(upper=ast.UnaryOp(
                                            op=ast.USub(),
                                            operand=ast.Constant(value=self._level),
                                        )),
                                        ctx=ast.Load(),
                                    ),
                                    op=ast.Add(),
                                    right=ast.List(elts=[ast.Constant(value=name.name)], ctx=ast.Load()),
                                )
                            ],
                            keywords=[],
                        ),
                        ast.Call(func=ast.Name(id='globals', ctx=ast.Load()), args=[], keywords=[]),
                        ast.Call(func=ast.Name(id='locals', ctx=ast.Load()), args=[], keywords=[]),
                        ast.List(elts=[], ctx=ast.Load()),
                        ast.Constant(value=0),
                    ],
                    keywords=[],
                )))

            # Rewrite "import abc.def" to "from .. import abc"
            if not name.asname:
                nodes.append(
                    ast.ImportFrom(
                        module=None,
                        names=[ast.alias(
                            name=name.name.split('.')[0],
                            asname=None,
                        )],
                        level=self._level + 1,
                    ))

            # Rewrite "import abc.def as xyz" to "from ..abc import def as xyz"
            else:
                nodes.append(
                    ast.ImportFrom(
                        module='.'.join(name.name.split('.')[:-1]),
                        names=[ast.alias(
                            name=name.name.split('.')[-1],
                            asname=name.asname,
                        )],
                        level=self._level + 1,
                    ))

        return nodes

    # pylint: disable=pylint(invalid-name)
    def visit_ImportFrom(self, node: ast.ImportFrom):
        node = self.generic_visit(node)

        # Rewrite "from abc import def (as xyz)" to "from ..abc import def (as xyz)"
        if node.level == 0 and node.module.split('.')[0] in self._modules:
            return ast.ImportFrom(
                module=node.module,
                names=node.names,
                level=self._level + 1,
            )

        return node

    # pylint: disable=pylint(invalid-name)
    def visit_Call(self, node: ast.Call):
        node = self.generic_visit(node)

        if isinstance(node.func, ast.Name) and isinstance(node.func.ctx, ast.Load):

            # rewrite __import__ calls
            if node.func.id == '__import__' and len(node.args) <= 5:

                # we do not support *x and **x
                has_starred = len([x for x in node.args if isinstance(x, ast.Starred)]) > 0
                has_kwargs_list = len([x for x in node.keywords if not x.arg]) > 0

                if not has_starred and not has_kwargs_list:

                    # extract arguments
                    arg_name = node.args[0] if len(node.args) > 0 else ast.Constant(value=None)
                    arg_globals = node.args[1] if len(node.args) > 1 else ast.Constant(value=None)
                    arg_locals = node.args[2] if len(node.args) > 2 else ast.Constant(value=None)
                    arg_fromlist = node.args[3] if len(node.args) > 3 else ast.List(elts=[])
                    arg_level = node.args[4] if len(node.args) > 4 else ast.Constant(value=0)

                    # extract keyword arguments
                    kwargs = {x.arg: x.value for x in node.keywords}
                    arg_name = kwargs['name'] if 'name' in kwargs else arg_name
                    arg_globals = kwargs['globals'] if 'globals' in kwargs else arg_globals
                    arg_locals = kwargs['locals'] if 'locals' in kwargs else arg_locals
                    arg_fromlist = kwargs['fromlist'] if 'fromlist' in kwargs else arg_fromlist
                    arg_level = kwargs['level'] if 'level' in kwargs else arg_level

                    # we support level 0 only
                    if isinstance(arg_level, ast.Constant) and arg_level.value == 0:

                        # transform name argument
                        arg_name = _transform_import_name_string(self._level, self._modules, arg_name)

                        # rewrite __import__ call
                        if arg_name:
                            return ast.Call(
                                func=node.func,
                                args=[arg_name, arg_globals, arg_locals, arg_fromlist, arg_level],
                                keywords=[],
                            )

            # rewrite import_module calls (importlib)
            if node.func.id == 'import_module' and len(node.args) <= 2:

                # we do not support *x and **x
                has_starred = len([x for x in node.args if isinstance(x, ast.Starred)]) > 0
                has_kwargs_list = len([x for x in node.keywords if not x.arg]) > 0

                if not has_starred and not has_kwargs_list:

                    # extract arguments
                    arg_name = node.args[0] if len(node.args) > 0 else ast.Constant(value=None)
                    arg_package = node.args[1] if len(node.args) > 1 else ast.Constant(value=None)

                    # extract keyword arguments
                    kwargs = {x.arg: x.value for x in node.keywords}
                    arg_name = kwargs['name'] if 'name' in kwargs else arg_name
                    arg_package = kwargs['package'] if 'package' in kwargs else arg_package

                    # transform name argument
                    arg_name = _transform_import_name_string(self._level, self._modules, arg_name)

                    # rewrite __import__ call
                    if arg_name:
                        return ast.Call(
                            func=node.func,
                            args=[arg_name, arg_package],
                            keywords=[],
                        )

        return node


def import_transform(source_path: Path, target_path: Path, level: int, modules: List[str]):

    # read file
    source = read_source(source_path)

    # transform
    tree = ast.parse(source, filename=str(source_path), type_comments=True)
    tree = ImportTransform(level, modules).visit(tree)
    tree = ast.fix_missing_locations(tree)
    target = ast.unparse(tree)

    # write file
    with open(target_path, 'w', encoding='utf-8') as tf:
        tf.write(target)



from os import walk
from os.path import join, relpath
from pathlib import Path
from zipfile import ZIP_DEFLATED, ZipFile


def zipit(source_path: Path, zip_path: Path, base_path: Path):

    # check pre-conditions
    assert not base_path.is_absolute(), 'zip base path is expected to be relative'

    # create zip file
    with ZipFile(zip_path, 'w', ZIP_DEFLATED) as handle:

        # zip directory
        if source_path.is_dir():
            for root, _, files in walk(source_path):
                for file in files:
                    path = join(root, file)
                    handle.write(path, join(base_path, relpath(path, source_path)))

        # zip individual file
        else:
            handle.write(str(source_path), str(base_path))



import sys

# avoid having the cwd in the path
# see https://docs.python.org/3/tutorial/modules.html#the-module-search-path
if '' in sys.path:
    sys.path.remove('')

from typing import List

from ppack.__main__ import main as main_pack
from pvariant.__main__ import main as main_variant
from pvendor.__main__ import main as main_vendor


def main(argv: List[str] = sys.argv[1:]):

    tool = argv[0] if len(argv) > 0 else None
    argv = argv[1:]

    if tool == 'pack':
        main_pack(argv)
    elif tool == 'variant':
        main_variant(argv)
    elif tool == 'vendor':
        main_vendor(argv)
    else:
        print('Usage: pdistx pack|variant|vendor --help')
        sys.exit(1)


if __name__ == '__main__':
    main()



import sys

# avoid having the cwd in the path
# see https://docs.python.org/3/tutorial/modules.html#the-module-search-path
if '' in sys.path:
    sys.path.remove('')

import argparse
from glob import glob
from os.path import join
from pathlib import Path
from traceback import print_tb
from typing import List

from pvariant.process import perform


def main(argv: List[str] = sys.argv[1:]):

    parser = argparse.ArgumentParser(prog='pvariant')

    parser.add_argument(
        '-d',
        dest='define',
        metavar='name[:type]=value',
        action='append',
        default=[],
        help='define variables to be replaced, e.g. -d __VARIANT__=PRO -d __LICENSE_CHECK__:bool=True',
    )

    parser.add_argument(
        '-f',
        dest='filter',
        metavar='filter',
        action='append',
        default=[],
        help='defines files and folders to be filtered out (glob pattern)',
    )

    parser.add_argument(
        '-z',
        dest='zip',
        metavar='zip',
        default=None,
        help='zip file path (target becomes relative path within zip file)',
    )

    parser.add_argument(
        'source',
        help='source path',
    )

    parser.add_argument(
        'target',
        help='target path (will be cleared)',
    )

    args = parser.parse_args(argv)

    definitions = {}

    for def_ in args.define:
        name, value = [*def_.split('=', 2), None][0:2]
        name, type_ = [*name.split(':', 2), 'str'][0:2]
        type_ = type_.lower()
        if type_ in ['', 'str']:
            value = value if value is not None else ''
        elif type_ == 'int':
            value = int(value if value is not None else 0)
        elif type_ == 'bool':
            value = value is not None and value.lower() == 'true'
        elif type_ == 'none':
            value = None
        else:
            raise ValueError('invalid definition type')
        if name:
            definitions[name] = value

    try:
        perform(
            Path(args.source),
            Path(args.target),
            definitions,
            [Path(path) for pattern in args.filter for path in glob(join(args.source, pattern), recursive=True)],
            Path(args.zip) if args.zip else None,
        )
    except Exception as ex:
        print(f'ERROR: {ex}')
        print_tb(ex.__traceback__)
        sys.exit(1)

    sys.exit(0)


if __name__ == '__main__':
    main()



# pylint: disable=import-outside-toplevel,deprecated-module,exec-used
def __pack_loader__():

    # import all required modules
    import imp
    import os.path
    import sys
    from collections import OrderedDict

    # pack data will be injected here
    pack_mode = ''
    pack_name = ''
    pack_modules = OrderedDict()

    # verify execution model
    if __name__ == '__main__' and pack_mode != 'main':
        raise RuntimeError('pack cannot be run as main script')

    if __name__ != '__main__' and pack_mode != 'package':
        raise RuntimeError('pack can only be run as main script')

    # transform this module to a package, if necessary
    if pack_mode == 'package':
        globals()['__path__'] = []
        globals()['__package__'] = __name__

    # set resource path
    resource_root = os.path.join(
        os.path.dirname(__file__),
        os.path.splitext(__file__)[0] + '_resources',
    )

    globals()['__resource__'] = os.path.join(
        resource_root,
        '__init__.py' if pack_mode == 'package' else '__main__.py',
    )

    # determine module base
    if pack_mode == 'main':
        base = pack_name
    else:
        base = __name__

    # clear all base hoisted modules from sys.modules
    for name in list(sys.modules.keys()):
        if (pack_mode == 'main' and name == base) or name.startswith(base + '.'):
            del sys.modules[name]

    # util: unqualify name
    def unqualify_name(fullname):
        if pack_mode == 'main' and fullname == base:
            return ''
        if fullname.startswith(base + '.'):
            return fullname[len(base) + 1:]
        return None

    # util: validate name
    def validate_name(name):
        return name is not None and name in pack_modules

    def assert_name(name):
        assert validate_name(name), 'unknown module'

    # util: get __file__ for module
    def get_dunder_file(fullname):
        name = unqualify_name(fullname)
        assert_name(name)
        _, is_package = pack_modules[name]
        return os.path.join(
            __file__,
            '/'.join(name.split('.') + (['__init__.py'] if is_package else [])),
        )

    # util: get __resource__ for module
    def get_dunder_resource(fullname):
        name = unqualify_name(fullname)
        assert_name(name)
        _, is_package = pack_modules[name]
        return os.path.join(
            resource_root,
            '/'.join(name.split('.') + (['__init__.py'] if is_package else [])),
        )

    # implement pack importer
    class PackImporter:

        # check if requested module can be handled
        def find_module(self, fullname, path=None):
            name = unqualify_name(fullname)
            if validate_name(name):
                return self
            return None

        def is_package(self, fullname):
            name = unqualify_name(fullname)
            assert_name(name)
            return pack_modules[name][1]

        def get_source(self, fullname):
            name = unqualify_name(fullname)
            assert_name(name)
            return pack_modules[name][0]

        def get_code(self, fullname):
            return compile(
                self.get_source(fullname),
                get_dunder_file(fullname),
                'exec',
            )

        def load_module(self, fullname):
            code = self.get_code(fullname)
            module = sys.modules.setdefault(fullname, imp.new_module(fullname))
            module.__file__ = get_dunder_file(fullname)
            setattr(module, '__resource__', get_dunder_resource(fullname))
            module.__loader__ = self
            if self.is_package(fullname):
                module.__path__ = []
                module.__package__ = fullname
            else:
                module.__package__ = fullname.rpartition('.')[0]
            exec(code, module.__dict__)
            return module

    sys.meta_path.insert(0, PackImporter())


__pack_loader__()



import sys

# avoid having the cwd in the path
# see https://docs.python.org/3/tutorial/modules.html#the-module-search-path
if '' in sys.path:
    sys.path.remove('')

import argparse
from glob import glob
from os.path import join
from pathlib import Path
from traceback import print_tb
from typing import List

from ppack.process import perform


def main(argv: List[str] = sys.argv[1:]):

    parser = argparse.ArgumentParser(prog='ppack')

    parser.add_argument(
        '-r',
        dest='resources',
        action='store_true',
        help='create a resources folder with all non-python files (it will be named <target>_resources and be cleared)',
    )

    parser.add_argument(
        '-m',
        dest='main',
        action='store_true',
        help='use __main__.py of the package as bootstrap code (default is to use the root __init__.py of the package)',
    )

    parser.add_argument(
        '-f',
        dest='filter',
        metavar='filter',
        action='append',
        default=[],
        help='defines files and folders to be filtered out (glob pattern)',
    )

    parser.add_argument(
        '-z',
        dest='zip',
        metavar='zip',
        default=None,
        help='zip file path (target becomes relative path within zip file)',
    )

    parser.add_argument(
        'source',
        help='source package path',
    )

    parser.add_argument(
        'target',
        help='target python (will be cleared)',
    )

    args = parser.parse_args(argv)

    try:
        perform(
            Path(args.source),
            Path(args.target),
            [Path(path) for pattern in args.filter for path in glob(join(args.source, pattern), recursive=True)],
            args.resources,
            args.main,
            Path(args.zip) if args.zip else None,
        )
    except Exception as ex:
        print(f'ERROR: {ex}')
        print_tb(ex.__traceback__)
        sys.exit(1)

    sys.exit(0)


if __name__ == '__main__':
    main()



from fnmatch import fnmatch
from pathlib import Path
from shutil import rmtree
from typing import List


def fnmatch_any(name: str, patterns: List[str]):
    for pattern in patterns:
        if fnmatch(name, pattern):
            return True
    return False


def rmpath(path: Path):
    if path.exists():
        if path.is_dir():
            rmtree(path)
        else:
            path.unlink()



from pathlib import Path
import re


def detect_source_encoding(path: Path):
    # according to PEP 0263
    # https://www.python.org/dev/peps/pep-0263/

    # read first two lines
    with open(path, 'rb') as file:
        lines = [file.readline(), file.readline()]

    # check if first line starts with a bom
    if lines[0].startswith(b'\xef\xbb\xbf'):
        return 'utf-8'

    # check for encoding marker
    for line in lines:
        m = re.match(rb'^[ \t\f]*#.*?coding[:=][ \t]*([-_.a-zA-Z0-9]+)', line)
        if m:
            return m.group(1).decode('latin-1')

    # default encoding
    return 'utf-8'


def read_source(path: Path):
    # detect encoding
    encoding = detect_source_encoding(path)

    # read all lines with proper encoding
    with open(path, 'r', encoding=encoding) as file:
        lines = file.read().split('\n')

    # remove encoding markers (see PEP 0263)
    for i in range(0, min(2, len(lines))):
        if re.match(r'^[ \t\f]*#.*?coding[:=]', lines[i]):
            lines[i] = ''

    # join lines and done
    return '\n'.join(lines)



import ast
from functools import reduce
from pathlib import Path

from pdistx.utils.source import read_source


class VariantTransform(ast.NodeTransformer):

    def __init__(self, definitions):
        self.definitions = definitions
        self._collect_used_definitions = None
        super().__init__()

    # pylint: disable=pylint(invalid-name)
    def visit_Assign(self, node: ast.Assign):
        node = self.generic_visit(node)

        if len(node.targets):
            target = node.targets[0]
            if isinstance(target, ast.Name) and isinstance(target.ctx, ast.Store) and target.id in self.definitions:
                return ast.Assign([target], ast.Constant(self.definitions[target.id]))

        return node

    # pylint: disable=pylint(invalid-name)
    def visit_Name(self, node: ast.Name):
        node = self.generic_visit(node)

        if isinstance(node.ctx, ast.Load) and node.id in self.definitions:
            if self._collect_used_definitions is not None:
                self._collect_used_definitions.add(node.id)
            return ast.Constant(self.definitions[node.id])

        return node

    # pylint: disable=pylint(invalid-name)
    def visit_BoolOp(self, node: ast.BoolOp):
        node = self.generic_visit(node)

        if reduce(lambda a, b: a and b, map(lambda v: isinstance(v, ast.Constant), node.values), True):
            values = map(lambda value: value.value, node.values)
            if isinstance(node.op, ast.And):
                return ast.Constant(reduce(lambda a, b: a and b, values, True))
            if isinstance(node.op, ast.Or):
                return ast.Constant(reduce(lambda a, b: a or b, values, False))

        return node

    # pylint: disable=pylint(invalid-name)
    def visit_UnaryOp(self, node: ast.UnaryOp):
        node = self.generic_visit(node)

        if isinstance(node.op, ast.Not) and isinstance(node.operand, ast.Constant):
            return ast.Constant(not node.operand.value)

        return node

    # pylint: disable=pylint(invalid-name)
    def visit_Compare(self, node: ast.Compare):
        node = self.generic_visit(node)

        if len(node.ops) == 1:
            op = node.ops[0]
            left = node.left
            right = node.comparators[0]

            if isinstance(left, ast.Constant) and isinstance(right, ast.Constant):
                if isinstance(op, ast.Eq):
                    return ast.Constant(left.value == right.value)
                if isinstance(op, ast.NotEq):
                    return ast.Constant(left.value != right.value)
                if isinstance(op, ast.Lt):
                    return ast.Constant(left.value < right.value)
                if isinstance(op, ast.LtE):
                    return ast.Constant(left.value <= right.value)
                if isinstance(op, ast.Gt):
                    return ast.Constant(left.value > right.value)
                if isinstance(op, ast.GtE):
                    return ast.Constant(left.value >= right.value)
                if isinstance(op, ast.Is):
                    return ast.Constant(left.value is right.value)
                if isinstance(op, ast.IsNot):
                    return ast.Constant(left.value is not right.value)

            if isinstance(left, ast.Constant) and isinstance(right, (ast.List, ast.Tuple)):
                if reduce(lambda a, b: a and b, [isinstance(child, ast.Constant) for child in right.elts], True):
                    right_values = [child.value for child in right.elts]
                    if isinstance(op, ast.In):
                        return ast.Constant(left.value in right_values)
                    if isinstance(op, ast.NotIn):
                        return ast.Constant(left.value not in right_values)

        return node

    # pylint: disable=pylint(invalid-name)
    def visit_If(self, node: ast.If):
        node = self.generic_visit(node)

        used_definitions = self._collect_used_definitions = set()
        self._collect_used_definitions = None

        if isinstance(node.test, ast.Constant):
            if node.test.value:
                body = node.body
            else:
                body = node.orelse
            if len(body) == 0:
                body = ast.Pass()
            return body

        if len(used_definitions) > 0:
            print('line {}: {} used in if statement, but code could not be reduced'.format(
                node.lineno, ', '.join(used_definitions)))

        return node


def variant_transform(source_path: Path, target_path: Path, definitions: dict):

    # read file
    source = read_source(source_path)

    # transform
    tree = ast.parse(source, filename=str(source_path), type_comments=True)
    tree = VariantTransform(definitions).visit(tree)
    tree = ast.fix_missing_locations(tree)
    target = ast.unparse(tree)

    # write file
    with open(target_path, 'w', encoding='utf-8') as tf:
        tf.write(target)
